from sqlalchemy import (
    Column,
    Integer,
    Text,
    Boolean,
    ForeignKey,
    Table)
from sqlalchemy.orm import relationship

import colander
import deform

from c2cgeoform.models import Base
from c2cgeoform.ext.deform_ext import RelationSelect2Widget


class EmploymentStatus(Base):
    __tablename__ = 'tests_employment_status'

    id = Column(Integer, primary_key=True)
    name = Column(Text, nullable=False)


class Phone(Base):
    __tablename__ = 'tests_phones'

    id = Column(Integer, primary_key=True, info={
        'colanderalchemy': {
            'widget': deform.widget.HiddenWidget()
        }})
    number = Column(Text, nullable=False, info={
        'colanderalchemy': {
            'title': 'Phone number'
        }})
    person_id = Column(Integer, ForeignKey('tests_persons.id'), info={
        'colanderalchemy': {
            'widget': deform.widget.HiddenWidget()
        }})
    verified = Column(Boolean)


class Tag(Base):
    __tablename__ = 'tests_tags'

    id = Column(Integer, primary_key=True)
    name = Column(Text, nullable=False)


person_tag = Table(
    "person_tag",
    Base.metadata,
    Column("tag_id", Integer, ForeignKey('tests_tags.id'),
           primary_key=True),
    Column("person_id", Integer, ForeignKey('tests_persons.id'),
           primary_key=True),
)


class Person(Base):
    __tablename__ = 'tests_persons'
    __colanderalchemy_config__ = {
        'title': 'A Person',
        'description': 'Tell us about you.'
    }
    __c2cgeoform_config__ = {
        'duplicate': True
    }

    id = Column(Integer, primary_key=True, info={
        'colanderalchemy': {
            'title': 'ID',
            'widget': deform.widget.HiddenWidget()
        }})
    hash = Column(Text, unique=True)
    name = Column(Text, nullable=False, info={
        'colanderalchemy': {
            'title': 'Your name',
        },
        'c2cgeoform': {
            'duplicate': True
        }})
    first_name = Column(Text, nullable=False, info={
        'colanderalchemy': {
            'title': 'Your first name'
        },
        'c2cgeoform': {
            'duplicate': True
        }})
    age = Column(Integer, info={
        'colanderalchemy': {
            'title': 'Your age',
            'validator': colander.Range(18,)
        }})
    phones = relationship(
        Phone,
        cascade="all, delete-orphan",
        info={
            'colanderalchemy': {
                'title': 'Phone numbers',
            }})
    tags = relationship(
        "Tag",
        secondary=person_tag,
        cascade="save-update,merge,refresh-expire",
        info={
            'colanderalchemy': {
                'title': 'Tags',
                'widget': RelationSelect2Widget(
                    Tag,
                    'id',
                    'name',
                    order_by='name',
                    multiple=True
                ),
                'includes': ['id']
            }})

    validated = Column(Boolean, info={
        'colanderalchemy': {
            'title': 'Validation',
            'label': 'Validated'
        }})



# coding=utf-8
import os
import sys
import transaction
from datetime import date, timedelta

from pyramid.paster import (
    get_appsettings,
    setup_logging,
    )

from pyramid.scripts.common import parse_vars


from ..models.meta import Base
from ..models import (
    get_engine,
    get_session_factory,
    get_tm_session,
    )

from ..models.c2cgeoform_demo import (
    schema,
    Address,
    BusStop,
    District,
    Excavation,
    Situation,
    ContactPerson
    )


def usage(argv):
    cmd = os.path.basename(argv[0])
    print('usage: %s <config_uri> [var=value]\n'
          '(example: "%s development.ini")' % (cmd, cmd))
    sys.exit(1)


def main(argv=sys.argv):
    if len(argv) < 2:
        usage(argv)
    config_uri = argv[1]
    options = parse_vars(argv[2:])
    setup_logging(config_uri)
    settings = get_appsettings(config_uri, options=options)

    engine = get_engine(settings)

    with engine.begin() as connection:
        init_db(connection, force='--force' in options)


def init_db(connection, force=False):
    if force:
        if schema_exists(connection, schema):
            connection.execute("DROP SCHEMA {} CASCADE;".format(schema))

    if not schema_exists(connection, schema):
        connection.execute("CREATE SCHEMA \"{}\";".format(schema))

    Base.metadata.create_all(connection)

    session_factory = get_session_factory(connection)

    with transaction.manager:
        dbsession = get_tm_session(session_factory, transaction.manager)
        setup_test_data(dbsession)


def schema_exists(connection, schema_name):
    sql = '''
SELECT count(*) AS count
FROM information_schema.schemata
WHERE schema_name = '{}';
'''.format(schema_name)
    result = connection.execute(sql)
    row = result.first()
    return row[0] == 1


def setup_test_data(dbsession):
    if dbsession.query(District).count() == 0:
        dbsession.add(District(id=0, name="Pully"))
        dbsession.add(District(id=1, name="Paudex"))
        dbsession.add(District(id=2, name="Belmont-sur-Lausanne"))
        dbsession.add(District(id=3, name="Trois-Chasseurs"))
        dbsession.add(District(id=4, name="La Claie-aux-Moines"))
        dbsession.add(District(id=5, name="Savigny"))
        dbsession.add(District(id=6, name="Mollie-Margot"))

    if dbsession.query(Situation).count() == 0:
        dbsession.add(Situation(id=0, name="Road", name_fr="Route"))
        dbsession.add(Situation(id=1, name="Sidewalk", name_fr="Trottoir"))
        dbsession.add(Situation(id=2, name="Berm", name_fr="Berme"))
        dbsession.add(Situation(
            id=3, name="Vegetated berm", name_fr=u"Berme végétalisée"))
        dbsession.add(Situation(id=4, name="Green zone", name_fr="Zone verte"))
        dbsession.add(Situation(id=5, name="Cobblestone", name_fr="Pavés"))

    if dbsession.query(BusStop).count() == 0:
        _add_bus_stops(dbsession)

    if dbsession.query(Address).count() == 0:
        dbsession.add(Address(id=0, label="Bern"))
        dbsession.add(Address(id=1, label="Lausanne"))
        dbsession.add(Address(id=2, label="Genève"))
        dbsession.add(Address(id=3, label="Zurich"))
        dbsession.add(Address(id=4, label="Lugano"))

    dbsession.flush()

    if dbsession.query(Excavation).count() == 0:
        for i in range(100):
            dbsession.add(_excavation(i, dbsession))


def _excavation(i, dbsession):
    situations = dbsession.query(Situation).all()
    addresses = dbsession.query(Address).all()

    contact = ContactPerson()
    contact.first_name = 'Leonard'
    contact.last_name = 'Michalon'

    excavation = Excavation(
        reference_number='ref{:04d}'.format(i),
        request_date=date.today() - timedelta(days=100 - i),
        description="Installation d'un réseau AEP",
        motif="Création d'un lotissement",
        location_district_id=0,
        location_street="48 avenue du Lac du Bourget",
        location_postal_code="73370",
        location_town="LE BOURGET DU LAC",
        # address_id=ForeignKey('c2cgeoform_demo.address.id'),
        # location_position = Column(geoalchemy2.Geometry('POINT'
        responsible_title="mr",
        responsible_name="Morvan",
        responsible_first_name="Arnaud",
        responsible_mobile="555-55555",
        responsible_mail="arnaud.morvan@camptocamp.com",
        responsible_company="Camptocamp",
        validated=True,
        # work_footprint=geoalchemy2.Geometry('MULTIPOLYGON'
        # photos = relationship(Photo,
    )
    for j in range(0, 3):
        excavation.situations.append(situations[(i + j) % len(situations)])
    excavation.contact_persons = [contact]
    excavation.address_id = addresses[i % len(addresses)].id
    return excavation


def _add_bus_stops(dbsession):
    """
    Load test data from a GeoJSON file.
    """
    import json
    from shapely.geometry import shape

    file = open(os.path.join(os.path.dirname(__file__),
                             '..',
                             'data',
                             'osm-lausanne-bus-stops.geojson'))
    geojson = json.load(file)
    file.close()

    bus_stops = []
    for feature in geojson['features']:
        id = feature['id'].replace('node/', '')
        geometry = shape(feature['geometry'])
        name = feature['properties']['name'] \
            if 'name' in feature['properties'] else ''
        bus_stop = BusStop(
            id=int(float(id)),
            geom='SRID=4326;' + geometry.wkt,
            name=name)
        bus_stops.append(bus_stop)

    dbsession.add_all(bus_stops)



import colander
from colanderalchemy import SQLAlchemySchemaNode
from sqlalchemy import and_, or_
from sqlalchemy.inspection import inspect


@colander.deferred
def deferred_request(node, kw):
    return kw.get('request')


@colander.deferred
def deferred_dbsession(node, kw):
    return kw.get('dbsession')


class GeoFormSchemaNode(SQLAlchemySchemaNode):

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.request = deferred_request
        self.dbsession = deferred_dbsession


class GeoFormManyToManySchemaNode(GeoFormSchemaNode):

    def __init__(self, class_, includes=None, *args, **kw):
        includes = [pk.name for pk in inspect(class_).primary_key]
        super().__init__(class_, includes, *args, **kw)

    def objectify(self, dict_, context=None):
        dbsession = self.bindings['dbsession']
        class_ = self.inspector.class_
        return dbsession.query(class_).get(dict_.values())


def manytomany_validator(node, cstruct):
    """
    Validate that cstruct values exists in related table.
    Note that entities are retrieved using only one query and placed in
    sqlalchemy identity map.
    """
    dbsession = node.bindings['dbsession']
    class_ = node.children[0].inspector.class_
    query = dbsession.query(class_)
    filters = []
    for dict_ in cstruct:
        filters.append(and_(*[getattr(class_, key) == value
                              for key, value in dict_.items()]))
    query = query.filter(or_(*filters))
    results = query.all()  # get all records in cache in one request
    diff = (set(tuple(dict_.values()) for dict_ in cstruct) -
            set([inspect(a).identity for a in results]))
    if len(diff) > 0:
        raise colander.Invalid(
            'Values {} does not exist in table {}'.
            format(", ".join(str(identity) for identity in diff),
                   class_.__tablename__))



import os
import unittest
from pyramid import testing
from pyramid.paster import get_appsettings
from sqlalchemy import engine_from_config
from webob.multidict import MultiDict

from c2cgeoform.models import (DBSession, Base)
from c2cgeoform.settings import apply_local_settings
from c2cgeoform import default_search_paths


class DatabaseTestCase(unittest.TestCase):
    def setUp(self):  # noqa
        curdir = os.path.dirname(os.path.abspath(__file__))
        configfile = os.path.realpath(
            os.path.join(curdir, '../../tests.ini'))
        settings = get_appsettings(configfile)
        apply_local_settings(settings)
        engine = engine_from_config(settings, 'sqlalchemy.')
        DBSession.configure(bind=engine)

        from .models_test import Person, EmploymentStatus, Tag  # noqa
        Base.metadata.create_all(engine)
        self.cleanup()

        # fill some test data into the `EmploymentStatus` and `Tags` table
        DBSession.add(EmploymentStatus(id=0, name='Worker'))
        DBSession.add(EmploymentStatus(id=1, name='Employee'))
        DBSession.add(EmploymentStatus(
            id=2, name='Self-employed and contractor'))
        DBSession.add(EmploymentStatus(id=3, name='Director'))
        DBSession.add(EmploymentStatus(id=4, name='Office holder'))

        DBSession.add(Tag(id=0, name='Tag A'))
        DBSession.add(Tag(id=1, name='Tag B'))
        DBSession.add(Tag(id=2, name='Tag C'))
        DBSession.add(Tag(id=3, name='Tag D'))
        DBSession.add(Tag(id=4, name='Tag E'))

        self.request = testing.DummyRequest(post=MultiDict(),
                                            dbsession=DBSession)
        testing.setUp(request=self.request)

        config = testing.setUp()
        config.add_route('form', '/{schema}/form/')
        config.add_route('view_user', '/{schema}/form/{hash}')
        config.add_route('confirm', '/{schema}/form/confirm')
        from deform import Form
        Form.set_zpt_renderer(default_search_paths)

    def tearDown(self):  # noqa
        self.cleanup()
        DBSession.remove()
        testing.tearDown()

    def cleanup(self):
        from .models_test import Person, EmploymentStatus, Phone, \
            Tag
        DBSession.query(Tag).delete()
        DBSession.query(Phone).delete()
        DBSession.query(Person).delete()
        DBSession.query(EmploymentStatus).delete()



from pyramid.view import view_config
from pyramid.view import view_defaults
from functools import partial

from sqlalchemy.orm import subqueryload

import colander
from c2cgeoform.schema import (
    GeoFormSchemaNode,
    GeoFormManyToManySchemaNode,
    manytomany_validator,
)
from c2cgeoform.ext.deform_ext import RelationCheckBoxListWidget
from c2cgeoform.views.abstract_views import AbstractViews, ListField

from ..models.c2cgeoform_demo import Excavation, Situation

_list_field = partial(ListField, Excavation)

base_schema = GeoFormSchemaNode(Excavation)
base_schema.add_before(
    'contact_persons',
    colander.SequenceSchema(
        GeoFormManyToManySchemaNode(Situation),
        name='situations',
        title='Situations',
        widget=RelationCheckBoxListWidget(
            Situation,
            'id',
            'name',
            order_by='name'
        ),
        validator=manytomany_validator
    )
)


@view_defaults(match_param='table=excavations')
class ExcavationViews(AbstractViews):

    _model = Excavation

    _list_fields = [
        _list_field('reference_number'),
        _list_field('request_date'),
        _list_field('description'),
        _list_field('location_town'),
        _list_field('responsible_company'),
        _list_field('situations',
                    renderer=lambda excavation: ", ".join(
                        [s.name for s in excavation.situations]),
                    filter_column=Situation.name)
    ]

    _id_field = 'hash'
    _base_schema = base_schema

    def _base_query(self):
        return self._request.dbsession.query(Excavation).distinct(). \
            join('situations'). \
            options(subqueryload('situations'))

    @view_config(route_name='c2cgeoform_index',
                 renderer='../templates/index.jinja2')
    def index(self):
        return super().index()

    @view_config(route_name='c2cgeoform_grid',
                 renderer='json')
    def grid(self):
        return super().grid()

    @view_config(route_name='c2cgeoform_item',
                 request_method='GET',
                 renderer='../templates/edit.jinja2')
    def edit(self):
        return super().edit()

    @view_config(route_name='c2cgeoform_item_action',
                 request_method='GET',
                 renderer='../templates/edit.jinja2')
    def duplicate(self):
        return super().duplicate()

    @view_config(route_name='c2cgeoform_item',
                 request_method='DELETE')
    def delete(self):
        return super().delete()

    @view_config(route_name='c2cgeoform_item',
                 request_method='POST',
                 renderer='../templates/edit.jinja2')
    def save(self):
        return super().save()



from pyramid.events import BeforeRender


def register_models(config, models):

    def table_pregenerator(request, elements, kw):
        if 'table' not in kw:
            kw['table'] = request.matchdict['table']
        return elements, kw

    def rec_with_pregenerator(route, pattern):
        config.add_route(route, pattern, pregenerator=table_pregenerator)

    tables = []
    for key, mapper in models:
        ca_config = getattr(mapper, '__colanderalchemy_config__', {})
        table = {}
        # key stand for url path, sometimes designated as table (in match dict)
        table['key'] = key
        table['title'] = ca_config.get('title', mapper.__tablename__)
        table['plural'] = ca_config.get('plural', mapper.__tablename__)
        tables.append(table)

    table_regex = '|'.join(['({})'.format(table['key']) for table in tables])
    base_route = '/{{table:{}}}'.format(table_regex)

    rec_with_pregenerator('c2cgeoform_index', '{}'.format(base_route))
    rec_with_pregenerator('c2cgeoform_grid', '{}/grid.json'.format(base_route))
    rec_with_pregenerator('c2cgeoform_item', '{}/{{id}}'
                          .format(base_route))
    rec_with_pregenerator('c2cgeoform_item_action', '{}/{{id}}/{{action}}'
                          .format(base_route))

    def add_global(event):
        event['tables'] = tables

    config.add_subscriber(add_global, iface=BeforeRender)
